##############################################################################
# This code is Copyright (c) 1996 Colin Howarth      www.muc.de/~colin
# All rights reserved. This program is free software; you may redistribute
# it and/or modify it under the same terms as Perl itself.
# $Id: PS.pm,v 0.13 1996/07/05 00:57:02 root Exp root $
##############################################################################


package PS;  ######################################################
BEGIN {

	($VERSION) = '$Revision: 0.13 $' =~ /([\d.]+)/ ;

	# The following is simply an aid to me to make things (slightly) quicker
	# I suggest getting rid of the following two lines, so that output goes
	# to STDOUT. This comment is only here in case I forget to delete this
	# for a release.

	open(PAGE, ">page.ps") || die "Can't open output file, page.ps";
	select PAGE;

	$, = " "; # this is the print separator, printed between LIST elements
	$\ = "\n"; # this is the print output record separator
}

END {
	close PAGE;
}

sub read_metrics {

# I should probably have Font as a class, no?

	my $font = shift;
	my %name2num;
	my $path = ($ENV{AFM}) ? $ENV{AFM} : "/usr/lib/X11/fonts/Type1Adobe/afm";

	return if (defined $font{$font});

	foreach $metrics (split /:/, $path) {
		if (open(AFM, "$metrics/$font")) {
			while(<AFM>) {
				if ( /^C (\d*) ; WX (\d*) ; N (.*) ; B (-?\d*) (-?\d*) (-?\d*) (-?\d*) ;/ ) {
					$font{$font}{Cwidth}[$1] = $2;
					$font{$font}{Box}[$1] = [$4,$5,$6,$7];
					$name2num{$3} = $1;
				}
				if ( /^KPX (.*) (.*) (-?\d*)/ ) {
					$font{$font}{Kern}{$name2num{$1}}{$name2num{$2}} = $3 if ($3 != 0);
				}
				if (/^FontBBox.* (\d*)$/) {
					$font{$font}{Height} = $1;
				}
			}
			close(AFM);
		}
	}
	if (!defined $font{$font}) {
		@{$font{$font}{Cwidth}} = (500) x 256;
		warn "Couldn't find font metrics for $font. Using default!";
	}
}

sub deref {

	my $ref = shift;
	my $refref = ref $ref;

	if   (!$refref)             { return $ref }
	elsif ($refref eq 'SCALAR') { return $$ref }
	elsif ($refref eq 'REF')    { return deref($$ref) }  # recursion
	elsif ($refref eq 'ARRAY')  { return $ref }  # but I want refs to arrays
	elsif ($refref eq 'HASH')   { return $ref }  #        ---- " ---
	elsif ($refref eq 'CODE')   { return deref(&$ref) }  # recursion
}


package Font;  #########################################################

# structure to store font info like metrics. Methods to manipulate
# fonts and return font-derived data.
# NOT IMPLEMENTED yet

sub new {
	my ($type, %params) = @_;
	my $self = {};

	bless $self, $type;
}

sub read {
	my @names = @_;
}

sub recode {
	my ($self, $encoding) = @_;

	$self->{Encoding} = $encoding;

# ? or provide direct access to font dict?

};

sub size_for {

# returns font size needed to render a character with exactly
# the given height
# eg. $size = $times_roman->size_for('a', 30)
# Extend to width ? Use wantarray?: ? Include offset from origin ?

	my ($self, $char, $size_wanted) = @_;

	$char = ord $char;
	my $char_height = $self->{Box}[$char][3];

# assumes chars sit on base-line! If you want REAL height the above needs
# to be my $char_height = self->{Box}[$char][3] - $self->{Box}[$char][1]

	return $size_wanted * 1000 / $char_height;
}


package Object;  #########################################################

sub new {
	my ($type, %params) = @_;
	my $self = {};
	
	$self->{Pos} = [0, 0];
					# position of this frame's origin within the parent frame
	$self->{Extent} = [0, 0];
					# width, height. absolute or as fraction of parent
	$self->{Origin} = [0, 0];  # [0,0] = bottom-left, [0,1] = top-left
					# relative origin, used for placement, scale & rotate
	$self->{Margin} = [0,0,0,0];
					# margin used for eg. text placement
	$self->{Borderwidth} = 0;
					# width of borderline in pt. Set to 0 for no border
	$self->{Rotate} = 0;
					# rotation angle in degrees about this object's origin
	$self->{Scale} = [1,1];
					# scale factor, [xsf, ysf], from 0 to 1
	$self->{Font} = 'Helvetica';
					# default font for any text that may be used
	$self->{Size} = 10;
					# default font size for any text that may be used
	$self->{Gsave} = 1;
					# whether graphics state should be unaffected by this
					# object. Objects which specifically want to change the
					# graphics state of the parent set this to 0.
	
	while (($key, $val) = each %params) {
		$self->{$key} = $val;
	}
	bless $self, $type;
}

sub render {
	my $self = shift;

	$self->prerender;
	if (defined $self->{PS}) { print "$self->{PS}"; }
	$self->postrender;
}

sub prerender {
	my $self = shift;
	my ($xr,$yr) = @{$self->{Origin}};
	my ($w,$h)= @{$self->get_extent};
	my ($x,$y) = @{$self->get_pos};
	my ($t,$sx,$sy) = ($self->{Rotate}, @{$self->{Scale}});

	my $gsave = $self->{Gsave} ? 'gsave' : '';

	print "% START object $self";
	
# translate origin to position given by Pos.

	print $gsave, $x, $y, "translate";

# now rotate about that point

	if ($t != 0) { print $t, "rotate" }

# and translate origin such that point given by Origin is on this point
# (but PS graphics state origin is at lower left corner)

	if ($sx != 1 || $sy != 1) { print $sx, $sy, "scale" }
	if ($xr != 0 || $yr != 0) { print -$w * $xr, -$h * $yr, "translate" }

	if (defined $self->{Background}) { $self->do_background; }
	
	if (defined $self->{Linewidth}) {
		print "$self->{Linewidth} setlinewidth";
	}
	
# '[3] 0'     = 3 on 3 off ...
# '[2 1] 0'   = 2 on 1 off 2 on 1 off ...
# '[3 5] 6'   = 2 off 3 on 5 off 3 on 5 off ...
# '[1 3 5] 0' = 1 on 3 off 5 on 1 off ...

	if (defined $self->{Dash}) {
		print "$self->{Dash} setdash";
	}
	
	if (defined $self->{Color}) {
		print setcolor($self->{Color});
	}
	
# don't stroke box if width = 0 . In PS width 0 means narrowest possible line

	if ($self->{Borderwidth}) {
		print "gsave $self->{Borderwidth} setlinewidth";
		print "0 0 $w $h rectstroke grestore";
	}

	if ($self->{Shadow}) {
		my($sh_color, $d) = @{$self->{Shadow}};

# $d = shadow width, $w, $h are frame width, height

		print "gsave", setcolor($sh_color),"[",$d,-$d,$w,$d,$w,0,$d,$h-$d,
		"] rectfill grestore";
	}
}

sub postrender {

	my $self = shift;
	my ($xr,$yr) = @{$self->{Origin}};
	my ($t,$sx,$sy) = ($self->{Rotate}, @{$self->{Scale}});
	my ($w,$h) = @{$self->get_extent};
	my ($x,$y) = @{$self->get_pos};

	my $grestore = $self->{Gsave} ? 'grestore' : '';

	foreach $object_ref (@{$self->{Objects}}) {
		$object_ref->render;
	}
	
	if ($xr != 0 || $yr != 0) { print $w * $xr, $h * $yr, "translate" }
	if ($sx != 1 || $sy != 1) { print 1/$sx, 1/$sy, "scale" }
	if ($t != 0) { print -$t, "rotate" }

	print -$x, -$y, "translate $grestore";
	print "% END object $self";
}

sub place {
	my $self = shift;
	my ($xr,$yr,$x,$y);
	
	if (@_ == 2) { ($x,$y) = @_; }
	if (@_ == 4) {
		($xr,$yr,$x,$y) = @_;
		$self->{Origin} = [$xr,$yr];
	}
	$self->{Pos} = [$x,$y];
	return $self;
}

sub extent {
	my $self = shift;
	my ($x,$y) = @_;

	$self->{Extent} = [$x,$y];
	return $self;
}
	
sub scale {
	my $self = shift;
	my ($sx,$sy) = @_;
	$self->{Scale} = [$sx,$sy];

	return $self;
}
	
sub rotate {
	my $self = shift;
	$self->{Rotate} = shift;
	return $self;
}

sub intersect {
	my ($self, $X,$Y) = @_;
	my ($w,$h) = @{$self->get_extent};
	my ($x,$y) = @{$self->get_pos};
	my ($xr,$yr) = @{$self->{Origin}};
	my ($dw,$dh) = ($w/2, $h/2);

# BUG: this ain't too hot on scaled/rotated objects :-(

	if ($X == 0) {
		$dw = 0;
	} else {
		my $T  = abs ($Y/$X);
		my $t  = $h/$w;
		
		if ( $T > $t ) { $dw = $dh / $T } else { $dh = $dw * $T }
	}

	if ($X > 0) { $x += $w/2 + $dw } else { $x += $w/2 - $dw }
	if ($Y > 0) { $y += $h/2 + $dh } else { $y += $h/2 - $dh }

	$x -= $xr * $w;
	$y -= $yr * $h;
	return [$x,$y];
}

sub join {
	my ($self,$start,$objs,@arrow) = @_;
	my ($xs,$ys,$X,$Y);

	my ($xo,$yo) = @{$start->get_pos};
	my ($xr,$yr) = @{$start->{Origin}};
	my ($wo,$ho) = @{$start->get_extent};

	if ($xr != 0 || $yr != 0) { $xo -= $wo * $xr; $yo -= $ho * $yr; }

	$xo += $wo/2; $yo += $ho/2;    # get object's centre

	foreach $obj (@$objs) {
		my ($xt,$yt) = @{$obj->get_pos};
		my ($xr,$yr) = @{$start->{Origin}};
		my ($wt,$ht) = @{$obj->get_extent};

		if ($xr != 0 || $yr != 0) { $xt -= $wt * $xr; $yt -= $ht * $yr; }

		$xt += $wt/2;               # get target's centre
		$yt += $ht/2;

		$X = $xt - $xo;             # get the join vector
		$Y = $yt - $yo;

		($xs,$ys) = @{$start->intersect($X,$Y)};
		($xt,$yt) = @{$obj->intersect(-$X,-$Y)};

		if (defined @arrow) {
			$self->arrow($xs,$ys,$xt-$xs,$yt-$ys,@arrow);
		} else {
			$self->line($xs,$ys,$xt,$yt);
		}
	}
}

sub get_pos {
	my $self = shift;
	my ($pw,$ph);
	my ($x,$y) = @{$self->{Pos}};

	if (defined $self->{Parent}) {
		($pw,$ph) = @{$self->{Parent}->get_extent};
	} else {
		($pw,$ph) = @{$self->get_extent};
	}

# for fractional x,y determine position as fraction of parent extent

	if (abs($x) <= 1) { $x *= $pw; }
	if (abs($y) <= 1) { $y *= $ph; }

# for -ve position determine position relative to top/right sides (a la X)

	if ($x < 0) { $x += $pw; }
	if ($y < 0) { $y += $ph; }

	return [$x,$y];
}

sub get_extent {
	my $self = shift;
	my ($w,$h,$pw,$ph);
	
	if (defined $self->{Parent}) {
		($pw,$ph) = @{$self->{Parent}->get_extent};
	} else {
		($pw,$ph) = @{$self->{Extent}};
	}

	if (defined $self->{Extent}) {
		($w,$h) = @{$self->{Extent}};
		if ($w <= 1) { $w *= $pw; }
		if ($h <= 1) { $h *= $ph; }
	} else {
		($w, $h) = ($pw, $ph);
	}
#	$w *= $self->{Scale}[0];
#	$h *= $self->{Scale}[1];

	return [$w, $h];
}

sub setcolor {

# input [r,g,b]       rgb are all [0-255] or all [0-1]
#       [-h,s,v]      h   is (0-360]   sv are [0-100]
#		  (g)           g   is [0-1]   (grey value. 0 = black, 1 = white)
# for those that don't know, HSV (hue saturation value) is a nice way of
# specifying colours. H is "colour", S is "pastel", V is "brightness"
		  
	my $color = shift;
	my ($c1,$c2,$c3);

	if (ref $color) {                          # if it's an (array) reference
		($c1, $c2, $c3) = @$color;
		if ($c1 < 0) {                     # if -ve means hsv. Use -360 for -0
			$c1 /= -360; $c2 /= 100; $c3 /= 100;
			return "$c1 $c2 $c3 sethsbcolor";
		} else {                           # otherwise it's rgb
		   if ( $c1 > 1 or $c2 > 1 or $c3 > 1) {        # if not fractional
				$c1 /= 255; $c2 /= 255; $c3 /= 255;       # then convert
			}
			return "$c1 $c2 $c3 setrgbcolor";
		}
	} else {                                   # else it's gray [0-1]
		return "$color setgray";
	}
}

sub label {
	my ($self, $text, $font, $size) = @_;
	my $child = $self->txt($text, $font, $size)->place(0,1.1, 0,0);
	$self->{Label} = $child;
	return $self;
}

sub title {
	my ($self, $text, $font, $size) = @_;
	my ($ml,$mr,$mt,$mb) = @{$self->{Margin}};

	my $child = $self->txt($text, $font, $size)->place(0.5,0.5, 0.5,-$mt/2);
	$self->{Title} = $child;
	return $self;
}

sub color {
	my ($self, $a, $b, $c) = @_;
	$self->{Color} = [$a, $b, $c];
	return $self;
}

sub b_color {
	my ($self, $a, $b, $c) = @_;
	$self->{Background} = [$a, $b, $c];
	return $self;
}

sub do_background {
	my $self = shift;
	my ($w,$h)= @{$self->get_extent};

	print "gsave", setcolor($self->{Background});
	print "0 0 $w $h rectfill grestore";
	return $self;
}

sub insert {
	my ($self, $child) = @_;

	$child->{Parent} = $self;
	push @{$self->{Objects}}, $child;
	return $child;
}

sub copy {
	my $self = shift;
	my $type = ref $self;
	my $copy = Object->new;
	my ($field, $val);
	
	while( ($field, $val) = each %$self ) {
		my $refref = ref $val;
		if ($refref eq 'REF')   { $val = deref $val }

		if    ($refref eq 'ARRAY')   { $val = [@$val] }
		elsif ($refref eq 'SCALAR')  { $val = \$$val }
		elsif ($refref eq 'HASH')    { $val = {%$val} }
		elsif ($refref eq 'CODE')    { $val = sub{ &$val } }  # ok ?
		else  { $val = $val->copy}

		$copy->{$field} = $val;
	}
	bless $copy, $type;
}

sub clone {
# See the section on cloning in doc!

	my $self = shift;
	my $type = ref $self;
	my $copy = Object->new;
	my ($field, $val);
	
	while( ($field, $val) = each %$self ) {
		$copy->{$field} = \$field;
	}
	bless $copy, $type;
}


package Path;  ###########################################################
@ISA = qw(Object);

sub new {
	my ($type, %params) = @_;
	my $self = new Object('PS' => 'newpath ', 'Name' => 'path', %params);

	bless $self, $type;
}

sub charpath {

# set $type to 'stroke' if you want to stroke the path
# ie. create an outline of the text
# ??? hmmm, don't understand this: true is ok for stroking as well?

	my($self, $text, $font, $size, $type) = @_;
	my $name = $self->{Name};
	my $bool;

	if (defined $type) {
		$bool = $type =~ /(fill)|(clip)/ ? 'true' : 'false';
	} else {
		$bool = 'true';
	}
	
	$self->{PS} = "/$font $size selectfont
	/$name {0 0 moveto ($text) $bool charpath} def";

	return $self;
}

sub clip {
	my $self = shift;
	my $name = $self->{Name};

	$self->{PS} .= " $name clip ";
	$self->{Gsave} = 0;
	return $self;
}

sub fill {
	my $self = shift;
	my $name = $self->{Name};

	$self->{PS} .= " $name fill ";
	return $self;
}

sub stroke {
	my ($self, $color) = @_;
	my $name = $self->{Name};

	if (defined $color) { $self->{PS} .= Object::setcolor($color); }
	$self->{PS} .= " $name stroke ";
	return $self;
}

sub moveto {
	my ($self, $x, $y) = @_;

	$self->{PS} .= " $x $y moveto ";
	return $self;
}
	
sub lineto {
	my ($self, $x, $y) = @_;

	$self->{PS} .= " $x $y lineto ";
	return $self;
}

# ?? add r(moveto|lineto) (r)curveto arc(n|t|to) rect(stroke|fill) etc ??


package Text;  ###########################################################
@ISA = qw(Object);

### BUG This is really slow and inefficient. Do it properly!

# Lsp    : line spacing as fraction of font size
# Psp    : paragraph spacing as fraction of font size
# Origin : default is top-left corner
# Pos    : default 1" in and down from top-left of parent
# Tabs   : default 10 tabs with 1" spacing

sub new {
	my ($type, %params) = @_;
	
	my $self = Object->new( Text => "", Font => 'Times-Roman',
	Size => '12', Lsp => '1.2', Psp => '2.0', Origin => [0, 1],
	Pos => [72, -72], Tabs => [map {$_ * 72} (1..10)], PS => [], %params);

	bless $self, $type;
}

sub Object::text {
	my ($self, %params) = @_;
	my ($px,$py) = @{$self->{Extent}};

	my $child = Text->new(Extent => [$px-144, $py-144], %params);

	$self->insert($child);
	return $child;
}

sub Object::txt {

# This function has been superseeded by Txt::new and will disappear shortly

	my ($self,$text,$font,$size) = @_;
	my $child = Text->new;

	$font = "Helvetica" if (!defined $font);
	$size = 12 if (!defined $size);
	PS::read_metrics($font);

	$child->{Text} = $text;
	$child->{Font} = $font;
	$child->{Size} = $size;
	$child->{Origin} = [$rx,$ry];
	$child->{Pos} = [$x,$y];
	$child->{Extent} = [1 + Text::width($font,$size,$text),
						$PS::font{$font}{Height} * $size / 1000];
	$self->insert($child);
	return $child;
}

sub render {
	my $self = shift;
	$self->prerender;                # this calls get_extent
	print "/mysave save def gsave";
	print "0 $self->{Height} translate";

	print join("\n", @{$self->{PS}});
                                 	# get_extent has to do the work anyway ...
	print "0 -$self->{Height} translate";
	print "grestore mysave restore";
	$self->postrender;
}

sub get_extent {
	my $self = shift;
	my ($ml,$mr,$mt,$mb) = @{$self->{Margin}};
	my ($w,$h)= @{$self->Object::get_extent};
	my $max_w = $w - ($ml + $mr);
	my $lines = 0;

	return [$w, $self->{Height}] if (defined $self->{Height});

	push @{$self->{PS}}, "/$self->{Font} $self->{Size} selectfont";
	push @{$self->{PS}}, "/size $self->{Size} def";

	if ($mt || $ml) { push @{$self->{PS}}, "$ml $mt neg translate"; }
	
	push @{$self->{PS}}, "0 -$self->{Size} translate\n0 0 moveto";
	
	if (exists $self->{File}) {
		open(REF, "$self->{File}") || die "Can't open $self->{File}";
		while(<REF>) {
			chop;
			$lines += $self->format($_, $max_w);
		}
	} else {
		foreach (split(/\n/, PS::deref($self->{Text}))) {
			$lines += $self->format($_, $max_w);
		}
	}

	if ($h == 0) { $h = $self->{Size} * ( 1 + ($lines - 1) * $self->{Lsp} ); }
	$self->{Height} = $h;
	return [$w, $h];
}

sub width {
	my ($font, $size, $word) = @_;

	$word = PS::deref($word);
	
	PS::read_metrics($font);       # hmmmmm

# cryptic comment or what? I only wrote it yesterday,
# and now even I don't know what it means :-(
# Does it mean one should 'register' fonts used at the beginning, so as not
# to need to check here *every* time?

	my @cwidth = @{$PS::font{$font}{Cwidth}};
	my %kern = %{$PS::font{$font}{Kern}};
	my ($c, $l, $width, $frag) = (0,0,0,'');

	foreach (split(//, $word )) {
		$c = ord $_;
		
		s/([\\\(\)])/\\$1/;        # PS treats \ ( ) as specials. quote.
		                           # $cwidth[$c] isn't defined for accented chars
		if (defined $cwidth[$c]) {
			$width += $cwidth[$c];
		} else {
			$width += 500;          # !!! add default width. Temporary kludge !!!
			warn "cwidth not defined for $_ ($c)" 
		}
		
		if ( $k = $kern{$l}{$c} ) {
			$width += $k;
			$frag .= ") $k M ($_";
		} else {
			$frag .= $_;
		}
		$l = $c;
	}
	$width *= ($size / 1000);

	return wantarray ? ($width, $frag) : $width;
}

sub tabify {
	my ($self, $x, $word) = @_;
	my @tabs = @{$self->{Tabs}};

	for $tab (@tabs) {
		if ($x <= $tab) { return $tab; } # return next TAB point
	}
	return $x;      # return current pos, if no more TABs defined
}

sub format {
	my $self = shift;
	my ($text, $linewidth) = @_;

## BUG what to do if linewidth is unknown, and passed as undef say?

	my $font = $self->{Font};
	my $size = $self->{Size};
	my $line_spacing = $self->{Lsp};
	my $para_spacing = $self->{Psp};
	my ($remain, $words, $frag, $width, @line) = ($linewidth,0,'',0, ());
	my ($lines, $x, $interword) = (1,0,0);
	my $space_width = width($font, $size, " ");
	
### BUG inifinite loop if width of word > linewidth

	while(defined $text and $text ne '') {
		($word, $sep, $text) = split(/([\s])/, $text, 2);
		($width, $frag) = width($font, $size, $word);

		if ($width <= $remain) {
			$words += 1;
			push @line, $frag;
			$x += $width;

			if (defined $sep) {
				if ($sep eq " ") {
					$x += $space_width;
				} else {
				if ($sep eq "\t") {
					my $xo = $x;
					$x = $self->tabify($x); 
					if ($x > $linewidth) {
						$x = $xo + $space_width;
					} else {
						push @line, ") show $x 0 moveto";
						push @{$self->{PS}}, "(" . join( " ) show (", @line);
						$words = 0; @line = (); 
					}
				}}
			}
			$remain = $linewidth - $x;
			
		} else {
			$text = defined($text) ? $word . $sep . $text : $word;

			# if word doesn't fit, put it back. If !defd $text ,
			# then it was last word
		
			$lines += 1;

			$interword = $words>1 ? ($remain + $space_width)/($words - 1) : 0;
			$interword *= (1000 / $size);
			
			push @{$self->{PS}}, "(" . join( " ) $interword M (" , @line) . ") show";
			push @{$self->{PS}}, "0 " . -$size*$line_spacing . " translate\n0 0 moveto";

			$words = 0; @line = (); $x = 0; $remain = $linewidth;
		}
	}
	
	push @{$self->{PS}}, "(" . join( " " , @line) . ") show";
	push @{$self->{PS}}, "0 " . -$size*$para_spacing . " translate\n0 0 moveto";

	return $lines;
}


package Txt;  ###########################################################
@ISA = qw(Object);

sub new {
	my ($type, %params) = @_;

	my $self = Object->new(Margin => [2,2,2,2], %params);

	bless $self, $type;
}

sub render {
	my $self = shift;

	($self->{Extent}, my $frag) = $self->get_extent;
	my($font, $size) = ( $self->{Font}, $self->{Size} );
	my($ml,$mr,$mt,$mb) = @{$self->{Margin}};
	
	$self->prerender;

	print "/size $size def /$font $size selectfont $ml $mb moveto ($frag) show";
	
	$self->postrender;
}

sub get_extent {
	my $self = shift;
	my $text = $self->{Text};
	my $font = $self->{Font};
	my $size = $self->{Size};
	my $height = $self->{Height};
	my($ml,$mr,$mt,$mb) = @{$self->{Margin}};

	my ($w, $frag) = Text::width($font, $size, $text);
	my $h = 0;

# get max character height from AFM font metrics info if {Height} not defined

	if (defined $height) {
		$h = $height;
	} else {
		for (split //, $text) {
			my $tmp = $PS::font{$font}{Box}[ord][3];
			if ( $tmp > $h ) { $h = $tmp; }
		}
		$h *= $size / 1000;      # adjust to real font size (ie. not 1000 pt)
	}

	$w += ( $ml + $mr);
	$h += ( $mt + $mb);

	return wantarray ? ([$w, $h], $frag) : [$w, $h] ;
}

sub Object::textbox {
	my ($self,$text,$font,$size,$height) = @_;
	 
	my $child = Txt->new(Text => $text, Font => ($font || "Helvetica"),
	                     Size => ($size || 12), Borderwidth => 1,
						 Height => $height);

	$self->insert($child);
}


package O_Txt;  ###########################################################
@ISA = qw(Circle Txt);

sub new {
	my ($type, %params) = @_;

	my $txt = Txt->new(%params);

#	passing %params to Txt->new AND to Circle->new is dubious.
#	Could cause problems. 

	my ($w, $h) = @{$txt->get_extent};

	$self = Circle->new(Radius => $w/2 * sqrt(2), %params)->scale(1, $h/$w);

# undo scale factor which turns circle into ellipse
	$self->{Text} = $txt->scale(1, $w/$h)->place(0.5,0.5, 0.5,0.5);
	
	$self->insert($txt);

	bless $self, $type;
}

sub b_color {
	my ($self, $a, $b, $c) = @_;
	$self->{Background} = [$a, $b, $c];
	return $self;
}

sub Object::o_txt {
	my ($self,$text,$font,$size) = @_;
	 
	my $child = O_Txt->new(Text => $text, Font => ($font || "Helvetica"),
	                       Size => ($size || 12));

	$self->insert($child);
}


package Img;  ###########################################################
@ISA = qw(Object);

sub new {
	my ($type, %params) = @_;
	my $self = Object->new( Scaletofit => 1, Maintainaspect => 1, %params);
	bless $self, $type;
}

sub render {
	my $self = shift;
	my ($w,$h)= @{$self->get_extent};
	my ($xs, $ys);
	
	$self->prerender;
	print "/mysave save def gsave";

	print "/showpage {} def /translate {pop pop} def";
	if (exists $self->{File}) {
		open(REF, "$self->{File}") || die "Can't open $self->{File}";
		while(<REF>) {
			chop; print;
			if ( /\%\%BoundingBox: (-?\d*) (-?\d*) (-?\d*) (-?\d*)/ ) {
				$xs = $w / ($3 - $1); $ys = $h / ($4 - $2);
				if ($self->{Scaletofit}) {
					if ($self->{Maintainaspect}) {
						print ($xs<$ys? "$xs $xs" : "$ys $ys", "scale");
					} else {
						print "$xs $ys scale";
					}
				}
				last;
			}
		}
		while(<REF>) { chop; print; }
	}
	print "grestore mysave restore";
	$self->postrender;
}

sub get_extent {
	my $self = shift;
	my ($w,$h) = @{$self->Object::get_extent};
	if ($w == 0 || $h == 0) {
		if (exists $self->{File}) {
			open(REF, "$self->{File}") || die "Can't open $self->{File}";
			while(<REF>) {
				if ( /\%\%BoundingBox: (-?\d*) (-?\d*) (-?\d*) (-?\d*)/ ) {
					if ($w == 0) {$w = ($3 - $1);}
					if ($h == 0) {$h = ($4 - $2);}
					last;
				}
			}
		}
	}
	return [$w,$h];
}


package Frame;  ######################################################
@ISA = qw(Object);

sub Object::frame {
	my ($self, $bw) = @_;
	my $child = Frame->new(Borderwidth => $bw);
	$self->insert($child);

	return $child;
}


package Data_Set;  #######################################################

# An object generally having Legend => text, Color => array ref, Labels =>
# array ref, Data => array ref (of scalars or array refs)

sub new {
	my ($type, %params) = @_;

	my $self = {};

	while (($key, $val) = each %params) {
		$self->{$key} = $val;
	}
	bless $self, $type;
}


package Graph;  ##########################################################
@ISA = qw(Object);

sub new {
	my ($type, %params) = @_;
	my $self = Object->new( Margin => [36,36,36,36], Font => 'Helvetica',
	Size => '12', Borderwidth => '1', Background => '1', %params);
	bless $self, $type;
}

sub label_x {
	my ($self, $text) = @_;
	my ($ml,$mr,$mt,$mb) = @{$self->{Margin}};

	my $child = $self->txt($text)->place(0.5,0.5, 0.5,$mb/3);
	$self->{Label_X} = $child;     # this is a Text object, so you can
	                               # change font, size, colour etc.
	return $self;
}

sub label_y {
	my ($self, $text) = @_;
	my ($ml,$mr,$mt,$mb) = @{$self->{Margin}};

	my $child = $self->txt($text)->place(0,0, $ml/5,0.5);
	$child->{Origin} = [0.5,0.5];
	$child->rotate(90);
	$self->{Label_Y} = $child;

	return $self;
}

sub minmax {
	my ($val, $min, $max);
	
	$min = $max = shift;  # initialise min and max with a real value
	foreach $val (@_) {
		if ($val < $min) { $min = $val }
		elsif ($val > $max) { $max = $val }
	}
	return ($min, $max);
}

sub split_labels_vals {
	my $self = shift;
	my @data = @{$self->{Data}};
	my (@vals, @labels, $data_sets);
	
	if (!ref($data[0])) {   # if the 1st value of data is not a reference
							# (to another array) then it's a 1-D data array
		@vals = @data;
	} else {
		$data_sets = @data;

		@labels = @{$data[0]};
		@vals = @{$data[1]};
	}
	return (\@vals, \@labels, $data_sets);
}

sub get_axis_vals {
	my ($self, $l, $max, $lab_sz) = @_;

	my $max_num_labels = int( 0.5 + $l / ($lab_sz * 2));
	my $label_interval = log($max / $max_num_labels) / log(10);
	                     # get log base 10 of the ideal inter-label interval	

	my $pow = int ($label_interval);
	my $lab_val = 10 ** ($label_interval - $pow);
	$pow = 10 ** $pow;
	                     # convert eg. 125 to 1.25 and 100	

# then increments are 1's, 2's, 5's, 10's, 20's, 50's, 100's etc.

	if ($lab_val <= 1) {
		$lab_val = 1;
	} elsif ($lab_val <= 2) {
		$lab_val = 2;
	} else {
		$lab_val = 5;
	}

	$lab_val *= $pow;
	my $num_labels = $max / $lab_val;
	if ($num_labels - int($num_labels) != 0) {$num_labels += 1}
# ie. if interval is 5 and max value is 23 then we want the final label = 25
	$num_labels = int($num_labels);

	return ($num_labels, $lab_val);
}


package Piechart;  ######################################################
@ISA = qw(Graph Object);

sub Object::piechart {
	my ($self, $data, $radius) = @_;
	my $child = Piechart->new;

	my ($ml,$mr,$mt,$mb) = @{$child->{Margin}};

	$child->{Data} = $data;
	$child->{Radius} = $radius;
	$child->{Extent} = [$radius*2 + $ml + $mr, $radius*2 + $mt + $mb];
	$self->insert($child);

	return $child;
}

sub render {
	my $self = shift;
	my($vals, $labels, $data_sets) = $self->split_labels_vals;
	my $radius = $self->{Radius};
	my ($grey,$sum,$i) = (0,0,0);
	my $delta_grey = 1/(@$vals - 1);
	my ($ml,$mr,$mt,$mb) = @{$self->{Margin}};

	$self->prerender;
	print $radius + $ml, $radius + $mb, "translate";

	print "/sector {/grey exch def /ang exch def newpath 0 0 moveto 0 0 $radius 0
	ang arc closepath grey setgray fill ang rotate} def";
	
	print "gsave";
	foreach (@$vals) { $sum += $_; }
	
	for($i = 0; $i < (@$vals); $i++) {
		if ($i % 2) {
			$grey = 1 - $delta_grey * ($i - 1)/2;
		} else {
			$grey = $delta_grey * $i/2;
		}
		print $vals->[$i]*360/$sum, "$grey sector";
	}
	print "grestore";

	print -($radius + $ml), -($radius + $mb), "translate";
	$self->postrender;
}


package Histogram;  ######################################################
@ISA = qw(Graph Object);

sub Object::histogram {
	my ($self, $data) = @_;

	my $child = Histogram->new(Data => $data, Bar_Color => 0.75);

	$self->insert($child);
	return $child;
}

sub render {
	my $self = shift;
	$self->prerender;

	my ($ml,$mr,$mt,$mb) = @{$self->{Margin}};
	my ($w,$h)= @{$self->get_extent};
	my ($width,$height) = ($w-$ml-$mr, $h-$mt-$mb);
	my ($font, $size) = ($self->{Font}, $self->{Size});

	my ($string, $iter, $i) = ('',0,0);
	my ($vals, $labels, $data_sets) = $self->split_labels_vals;
	my ($min, $max) = Graph::minmax(@$vals);
	my ($num_labels, $lab_val) = $self->get_axis_vals($height, $max, $size);
	my $columns = @$vals;
	my $cwidth = $width / $columns;
	my $vscale = $height / ($num_labels * $lab_val);

	foreach $val (@$vals) {
		$string .= $ml + 5 + $cwidth * $iter++ . " $mb " . ($cwidth - 5) ." ". $val * $vscale .  " ";
	}

	print "gsave", Object::setcolor($self->{Bar_Color});
	print "[ $string ] rectfill";
	print "grestore";
	
	print "$ml $h $mt sub moveto $ml $mb lineto $w $mr sub $mb lineto stroke";

	$iter = 0;
	foreach $label (@$labels) {
		$self->txt($label, $font, $size)->place(0.5, 0.1,
		                      $ml + 5 + $cwidth * $iter++ + ($cwidth - 5)/2,
									 2*$mb/3);
	}

	for ($i = 0; $i <= $num_labels; $i++) {
		$y_label = $i * $lab_val;
		$self->txt($y_label, $font, $size)->place(1, 0.5,
		                                    $ml - 5,
														$mb + $y_label * $vscale);
	}

	$self->postrender;
}


package Scatter;  ######################################################
@ISA = qw(Graph Object);

sub Object::scatter {
	my ($self, $data_sets, %params) = @_;
	my $child = Scatter->new(Data => $data_sets, %params);
	$self->insert($child);

	return $child;
}

sub render {
	my $self = shift;
	my ($ml,$mr,$mt,$mb) = @{$self->{Margin}};
	my ($w,$h)= @{$self->get_extent};
	my ($width,$height) = ($w-$ml-$mr, $h-$mt-$mb);
	my ($font, $size) = ($self->{Font}, $self->{Size});
		
	my $shape = $self->{Shape};
	my ($minx,$maxx,$miny,$maxy) = (0,0,0,0);
	my ($data_set, $i);

	foreach $d_set (@{$self->{Data}}) {
		my $x_coords = PS::deref $d_set->{Values}->[0];
		my $y_coords = PS::deref $d_set->{Values}->[1];

		($minx, $maxx) = Graph::minmax($minx, $maxx, @$x_coords);
		($miny, $maxy) = Graph::minmax($miny, $maxy, @$y_coords);
	}
	
	my $x_label_width = Text::width($font, $size, $maxx);
	
	my ($num_labels_x, $lab_val_x) =
					$self->get_axis_vals($width, $maxx, $x_label_width);
	my ($num_labels_y, $lab_val_y) =
					$self->get_axis_vals($height, $maxy, $size);
					
	my $hscale = $width / ($num_labels_x * $lab_val_x);
	my $vscale = $height / ($num_labels_y * $lab_val_y);

	$self->prerender;

	print "/$font $size selectfont";
	print "$ml $h $mt sub moveto $ml $mb lineto $w $mr sub $mb lineto stroke";

	for ($i = 0; $i <= $num_labels_x; $i++) {
		$x_label = $i * $lab_val_x;
		$self->txt($x_label, $font, $size)->place(0.5, 1,
		                                    $ml + $x_label * $hscale,
														$mb - 5);
	}

	for ($i = 0; $i <= $num_labels_y; $i++) {
		$y_label = $i * $lab_val_y;
		$self->txt($y_label, $font, $size)->place(1, 0.5,
		                                    $ml - 5,
														$mb + $y_label * $vscale);
	}

	print "$ml $mb translate";
	foreach $d_set (@{$self->{Data}}) {
		my $x_coords = PS::deref $d_set->{Values}->[0];
		my $y_coords = PS::deref $d_set->{Values}->[1];

		print "gsave";
#		print "/shape {-2 -2 4 4 rectfill} def";
		print "/shape {";
		$d_set->{Shape}->render;
		print "} def";
		while(  $x = shift @$x_coords ) {
			$y = shift @$y_coords;
			$x *= $hscale;
			$y *= $vscale;
			print "$x $y translate shape $x neg $y neg translate";
		}
		
		print "stroke grestore";
	}
	print "-$ml -$mb translate";
	$self->postrender;
}


package Line;  ######################################################
@ISA = qw(Object);

sub Object::line {
	my ($self, $x, $y, $x2, $y2) = @_;
	my $child = Line->new;

	$child->{PS} = "$x $y moveto $x2 $y2 lineto stroke";
	$self->insert($child);
	return $child;
}


package Arrow;  ######################################################
@ISA = qw(Object);

sub Object::arrow {
	my ($self, $x, $y, $dx, $dy, $type, $text, $font, $size) = @_;
	my $child = Arrow->new;

	my ($l, $a) = (sqrt($dx*$dx + $dy*$dy), (90/atan2(1,0)) * atan2($dy, $dx));
	my ($hdx, $hdy, $hc, $line_width) = (10, 3, 4, 0.5);

	$child->{Delta} = [$dx,$dy];
	$child->{Length} = $l;
	$child->{Rotate} = $a;
	$child->{Type} = $type;            # f: filled, d: double-headed, fd: both
	$child->{Head} = [$hdx, $hdy, $hc];
	$child->{Pos} = [$x,$y];

	my $line_width = $child->{Line_width} = $line_width;

	if ( $text ) {
		if (!$font) {$font = $child->{Font}};
		if (!$size) {$size = $child->{Size}};

		if ( $child->{Type} =~ /d/ ) { $hdx = 0 }

		my $txt = $child->txt($text, $font, $size)
		                ->place(0.5,-0.1, ($l-$hdx)/2, 0);

# bug: changing font, size or text dosn't change Extent.
# doing this with auto-derefs looks hairy... what we need here is a *concept*
# [sigh] it's probably called OOP :-(

		$child->{Text} = $txt;
	}

	$self->insert($child);
	return $child;
}

sub render {
	my $self = shift;
	$self->prerender;

	my $l = $self->{Length};
	my ($hdx, $hdy, $hc) = @{$self->{Head}};
	my $line_width = $self->{Line_width};

	# hdx is distance back down, hdy distance perpendicular to shaft.
	# hc is "indent" distance back towards head.
	# Ie hc = hdx gives ->, hc = 0 gives triangular head.
	# try 0 and -ve vals for |---| or >----<

	my $fill = ( $self->{Type} =~ /f/ ) ?  "fill" : "" ;
	
	print "gsave $line_width setlinewidth 0 setlinejoin $l 0 moveto $hdx neg $hdy
	rlineto $hc $hdy neg rlineto $hc neg $hdy neg rlineto closepath";
	
	if ( $self->{Type} =~ /d/ ) {
		print "$fill 0 0 moveto $hdx $hdy rlineto $hc neg $hdy neg rlineto
		$hc $hdy neg rlineto closepath $fill $hdx $hc sub 0 moveto $l $hdx
		$hc sub sub 0 lineto stroke grestore";
	} else {
		print "$fill 0 0 moveto $l $hdx $hc sub sub 0 lineto stroke grestore";
	}
		
	$self->postrender;
}


package Grid; #########################################################
@ISA = qw(Object);

sub Object::grid {

# Here, the meanings of Pos Origin and Extent are subverted. Origin is,
# well, the origin.

	my ($self, $sx, $sy, $msx, $msy, $mmsx, $mmsy) = @_;
	my $child = Grid->new;

	$msx = 0 if (!$msx); $mmsx = 0 if (!$mmsx);
	$msy = 0 if (!$msy); $mmsy = 0 if (!$mmsy);

	$child->{Intervals} = [$sx, $sy, $msx, $msy, $mmsx, $mmsy];

	$self->insert($child);
	return $child;
}

sub render {
	my $self = shift;

	my ($pw, $ph) = @{$self->{Parent}->get_extent};
	my ($sx, $sy, $msx, $msy, $mmsx, $mmsy) = @{$self->{Intervals}};
	my ($xr, $yr) = @{$self->{Origin}}; 

	$self->{Pos} = $self->{Origin}; 

	my $cx = $pw * $xr;
	my $cy = $ph * $yr;

	$self->prerender;

	print "/verticals { newpath { 0 moveto 0 $ph rlineto } for stroke } def
	/horizontals { newpath { 0 exch moveto $pw 0 rlineto } for stroke } def

	gsave 0 setlinewidth";

	if ($mmsx) { print "$cx $mmsx $pw verticals   $cx -$mmsx 0 verticals" }
	if ($mmsy) { print "$cy $mmsy $ph horizontals $cy -$mmsy 0 horizontals" }

	print "0.25 setlinewidth";

	if ($msx) { print "$cx $msx $pw verticals   $cx -$msx 0 verticals" }
	if ($msy) { print "$cy $msy $ph horizontals $cy -$msy 0 horizontals" }

	print "0.5 setlinewidth";
	
	if ($sx) { print "$cx $sx $pw verticals   $cx -$sx 0 verticals" }
	if ($sy) { print "$cy $sy $ph horizontals $cy -$sy 0 horizontals" }

	print "grestore";

	$self->postrender;
}


package Bezier; #########################################################
@ISA = qw(Object);

sub Object::bezier {
	my ($self, @args) = @_;
	my $child = Bezier->new;

	$child->{PS} = "@args[0..1] moveto @args[2..7] curveto stroke";

	$self->insert($child);
	return $child;
}


package Circle; #########################################################
@ISA = qw(Object);

sub Object::circle {

	my ($self, $x,$y,$r) = @_;
	my $child = Circle->new(Radius => $r, Pos => [$x, $y]);;

	$self->insert($child);
	bless $child;
}

sub new {
	my($type, %params) = @_;
	my $self = Object->new(Origin, [0.5, 0.5], %params);

	$self->{Extent} = [$self->{Radius}*2, $self->{Radius}*2];

	bless $self, $type;
}

sub render {
	my $self = shift;

	my $r = $self->{Radius};
	my $back = $self->{Background};
	$self->prerender;

# need to gsave/restore around stroke because stroke clears the path
# which we may want to fill

	print "newpath $r $r $r 0 360 arc gsave stroke grestore";

	if (defined $back) {
		print "gsave", Object::setcolor($back), "fill grestore";
	}

	$self->postrender;
}	

sub do_background {
	my $self = shift;
	# do nothing. Stops filling of extent rectangle
}

sub intersect {

# overrides base method, cos a circle isn't very rectangular

# if this is confusing draw an ellipse with
# 'a' as the "x-axis radius"
# 'b' as the "y-axis radius"
# [X,Y] as the vector from the centre to a point outside the ellipse
# and remember that for a circle   (x/r)**2 + (y/r)**2 = (r/r)**2 = 1
# and          that for an ellipse (x/a)**2 + (y/b)**2 = (r/r)**2 = 1

	my ($self, $X,$Y) = @_;
	my ($x,$y) = @{$self->get_pos};
	my ($sx,$sy) = @{$self->{Scale}};
	my ($w,$h) = @{$self->get_extent};

	$w *= $sx;   # the following calcs need the *real* height and width
	$h *= $sy;

	my $r = $self->{Radius};
	my ($a, $b) = ($r * $sx, $r * $sy);

	my $H = sqrt($X*$X + $Y*$Y);
	my $R =  $H / sqrt( ($X/$a)**2 + ($Y/$b)**2 );

	$x += $a + $R * $X/$H;   $y += $b + $R * $Y/$H ;

#	$x += $r + $r*$X/$H; $y += $r + $r*$Y/$H ;
	
	$x -= $self->{Origin}->[0] * $w;
	$y -= $self->{Origin}->[1] * $h;
	return [$x,$y];
}


package Rbox; #########################################################
@ISA = qw(Object);

sub Object::rbox {
	my ($self, $x0, $y0, $x1, $y1, $r) = @_;
	my $child = Rbox->new;
	
	$child->{PS} = "/fourpops { 4 {pop} repeat } def
	newpath $x0 $y0 $r add moveto
	$x0 $y1 $x0 $r add $y1 $r arcto fourpops
	$x1 $y1 $x1 $y1 $r sub $r arcto fourpops
	$x1 $y0 $x1 $r sub $y0 $r arcto fourpops
	$x0 $y0 $x0 $y0 $r add $r arcto fourpops stroke";

	$self->insert($child);
	return $child;
}


package Rule; #########################################################
@ISA = qw(Object);

sub Object::rule {
	my ($self, $x, $y, $w, @mwidth) = @_;
	my $child = Rule->new;
	my $x2 = $x + $w;

	$child->{PS} = "newpath $x $y moveto $x2 $y lineto stroke\n";
	$child->{PS} .= "/$self->{Font} $self->{Size} selectfont\n";

	my $mark_height = 6;
	my @mark_val = (0..20);
	my $idx = 0;

	$mark_inc = shift (@mwidth);
	for ($mark = $x; $mark < ($x + $w); $mark += $mark_inc) {
		$child->{PS} .= "$mark $y moveto 0 $mark_height rlineto stroke\n";
		$child->txt($mark_val[$idx++])
		      ->place(0.5, 0, $mark, $y + 8);
	}
	$mark_height *= 0.5;

	foreach $mark_inc (@mwidth) {
		for ($mark = $x; $mark < ($x + $w); $mark += $mark_inc) {
			$child->{PS} .= "$mark $y moveto 0 $mark_height rlineto stroke\n";
		}
		$mark_height *= 0.5;
	}

	$self->insert($child);
	return $child;
}


package Page;  ##########################################################
@ISA = qw(Object);

sub new {
	my ($type, %params) = @_;
	my $self = Object->new( Extent, [595, 842], %params );
	# 595x842 pts = A4
	# I'm not telling you what Legal and Letter are in points :-)
	# hmmm. maybe should set {Paper_Type} according to locale? 

	bless $self, $type;
}

### BUG (?) this is only any good for single pages. Fix it.

sub showpage {
	my $self = shift;
	$date = `date`; chop $date;
	$user = `whoami`; chop $user; $user .= "@" . `hostname`; chop $user;

	# haven't used here docs, cos here docs is a silly name.

	print "%!PS-Adobe-3.0\n%%Creator: PerlScript\n%%CreationDate: $date";
	print "%%For: $user\n%%EndComments\n%%Beginprolog";
	print "%%BeginResource: perlPS-Prolog 1.0 0";
	print "/Times-Roman findfont dup length dict begin {1 index /FID ne
	{def} {pop pop} ifelse} forall /Encoding ISOLatin1Encoding def
	currentdict end /Times-Roman exch definefont pop";
	print "/M {exch show size mul 1000 div 0 rmoveto} def";
	print "%%EndResource\n%%Page: 1 1\n%%BeginPageSetup\n0.2 setflat";

	$self->render;

	print "showpage\n%%EndPageSetup\n%%PageTrailer\n%%Trailer\n%%EOF";
}

return 1;
