#!/usr/local/bin/perl5
use Pathop;

$| = 1;
#
# make test files for cshpath
#
foreach $dir ('test', 'test/xlib', 'test/ylib', 'test/zlib') {
	if (! -d $dir) {
		mkdir($dir, 0777) || die("cannot make test directory.");
	}
}
foreach $file ('test/.null', 'test/xlib/null1', 'test/ylib/null2', 'test/zlib/null3') {
	if (! -f $file) {
		open(FILE, ">$file") || die("cannot make test files.");
		close(FILE);
	}
}
open(LOG, ">LOG") || die("cannot make log file");
print "PATHOP TEST START.";
$errcount = 0;
#
# exppath
#
%data = (
	# INPUT				   OUTPUT
	'/t1/{t2,t3}/t4'		=> '/t1/t2/t4 /t1/t3/t4',
	'/t1/{,t2}/t3'			=> '/t1//t3 /t1/t2/t3',
	'/t1/{}t2/t3'			=> '/t1/t2/t3',
	'/t1/{t2/t3,t4/t5}/t6'		=> '/t1/t2/t3/t6 /t1/t4/t5/t6',
	'/{t1,t2,t3}/{t4,t5}t6/{t7,t8}X'=> '/t1/t4t6/t7X /t1/t4t6/t8X /t1/t5t6/t7X /t1/t5t6/t8X /t2/t4t6/t7X /t2/t4t6/t8X /t2/t5t6/t7X /t2/t5t6/t8X /t3/t4t6/t7X /t3/t4t6/t8X /t3/t5t6/t7X /t3/t5t6/t8X',
);
foreach $in (keys(%data)) {
	@out = exppath($in);
	if ("@out" eq $data{$in}) {
		print ".";
	} else {
		$errcount++;
		print "X";
		print LOG "<<<ERROR>>>\n";
		print LOG "Result: @out = exppath($in);\n";
		print LOG "Expected: $data{$in} = exppath($in);\n";
	}
}
#
# globpath
#
%data = (
	# INPUT			   OUTPUT
	'test/*ull'		=> '',
	'test/.*ull'		=> 'test/.null',
	'test/.null'		=> 'test/.null',
	'test/?lib/null?'	=> 'test/xlib/null1 test/ylib/null2 test/zlib/null3',
	'test/*/*1'		=> 'test/xlib/null1',
	'test/*/null[12]'	=> 'test/xlib/null1 test/ylib/null2',
	'test/*/null[1-2]'	=> 'test/xlib/null1 test/ylib/null2',
);
foreach $in (keys(%data)) {
	@out = globpath($in);
	if ("@out" eq $data{$in}) {
		print ".";
	} else {
		$errcount++;
		print "X";
		print LOG "<<<ERROR>>>\n";
		print LOG "Result: @out = globpath($in);\n";
		print LOG "Expected: $data{$in} = globpath($in);\n";
	}
}
#
# cshpath
#
%data = (
	# INPUT				   OUTPUT
	'test/{?lib,z*b}/null[23]'	=> 'test/ylib/null2 test/zlib/null3 test/zlib/null3',
);
foreach $in (keys(%data)) {
	@out = cshpath($in);
	if ("@out" eq $data{$in}) {
		print ".";
	} else {
		$errcount++;
		print "X";
		print LOG "<<<ERROR>>>\n";
		print LOG "Result: @out = cshpath($in);\n";
		print LOG "Expected: $data{$in} = cshpath($in);\n";
	}
}
close(LOG);
if ($errcount) {
	print "FAILED. $errcount errors occured.\n";
	open(LOG, "LOG") || die("cannot open LOG file.");
	@log = <LOG>;
	print @log;
	close(LOG);
	exit(1);
}
print "COMPLETED.\n";
exit(0);
