/* snarf, the Simple Non-interactive All-purpose Resource Fetcher
** Copyright (C) 1995 Zachary Beane
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public LIcense as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is dsitributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILIY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** The author of this program may be reached via email at
** xach@mint.net or via USPS at 17 Talmadge Rd., Waite, ME 04492, USA.
**
**   $Id: url.c,v 1.7 1998/02/25 15:40:32 xach Exp $
**   $Source: /usr/local/cvs/snarf/url.c,v $
*/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "protocol.h"
#include "url.h"


struct url parse_url(char *string)
{
  struct url u;
  register int i, sp;
  int slashcount;
  char tmp[4096];

  i = sp = 0;

  if(! isurl(string) ) {
    fprintf(stderr, "%s: not an URL that I understand\n", string);
    exit(1);
  }
    


/* get the service type */
  for(i = 0; string[sp] != ':'; i++, sp++){
    tmp[i] = string[sp];
  }
  
  tmp[i] = '\0';

  u.u_service = strdup(tmp);

  /* this skips the :// section of string */

  while(!isalnum(string[sp++]))
    /* do nothing */; 

  --sp; 


  /* DWD kludge to handle usernames and passwords */

  i = 0;
  if(zmatch("?://?:?@?", string)) {
    char mytmp[1024];
    int t = 0;		/* temp index */

    i = sp;
    
    while((string[i] != ':') 
	  && (string[i] != '@') 
	  && (string[i] != '\0')
	  && (t < 1024)) 
    {
      mytmp[t++] = string[i++];
    }
    mytmp[t] = '\0';

    u.u_user = strdup(mytmp);

    if(string[i] == '@') {
      u.u_pass = strdup("");
    }
    else {
      t = 0;
      i++;
      while((string[i] != '\0')
	    && (string[i] != '@')
	    && (t < 1024))
      {
	mytmp[t++] = string[i++];
      }
      
      mytmp[t] = '\0';
      u.u_pass = strdup(mytmp);
    }
  }
  else {
      u.u_user = strdup("anonymous");
      u.u_pass = strdup("snarf@");
  }

  if(i > 0) 
    sp = ++i;
   
/* get the hostname */
  for(i = 0; string[sp] != '/' && string[sp] != ':' 
	&& string[sp] != '\0'; i++, sp++){
    tmp[i] = string[sp];
  }
  
  tmp[i] = '\0';
  u.u_host = strdup(tmp);

/* if it has a :<port> in the name, parse it */
  if(string[sp] == ':'){
    sp++;
    i = 0;
    while(isdigit(string[sp]))
      tmp[i++] = string[sp++];
    tmp[i] = '\0';
    u.u_port = (unsigned int) atoi(tmp);
  }
  else
    u.u_port = getport(u.u_service);

  /* get the number of slashes so we can get the document and pathname */
  for(i = sp, slashcount = 0; string[i] != '\0'; i++)
    if(string[i] == '/') 
      slashcount++;

  /* get the pathname */
  for(i = 0; slashcount > 0; i++, sp++){
    tmp[i] = string[sp];
    if(string[sp] == '/')
      slashcount--;
  }

  tmp[i] = '\0';

  if(tmp[0] == '\0')
    u.u_path = strdup("/");
  else 
    u.u_path = strdup(tmp);




    

/* get the document name */
  for(i = 0; string[sp] != '\0'; i++, sp++)
    tmp[i] = string[sp];
  tmp[i] = '\0';

  u.u_file = strdup(tmp);

  return(u);
}


int isurl(char *string)
{
  return(zmatch("?://?", string));
}


int zmatch(const char *match, char *string)
{
  register unsigned int sp = 0;
  register unsigned int mp = 0; /* string pointer and match pointer */
  int tolerance = 30000;
  int unmatched = 0;
  
  while(string[sp] != '\0'){
    if(match[mp] == '?'){
      mp++;
      sp++;
      tolerance = 30000;
      continue;
      unmatched = 0;
    }

    if(string[sp] == match[mp]){
      mp++;
      sp++;
      tolerance = 0;
      continue;
      unmatched = 0;
    }

    unmatched++;
    sp++;
  }
  
  if(mp == strlen(match))
    return(1);
  else
    return(0);
}

/*\
This function returns default port numbers for the supported services,
or zero otherwise.
\*/

int getport(char *string)
{
  if((strcmp("http",string)) == 0)
    return(HTTP);
  if((strcmp("ftp",string)) == 0)
    return(FTP);
  if((strcmp("finger",string)) == 0)
    return(FINGER);
  if((strcmp("gopher",string)) == 0)
    return(GOPHER);
  return 0;
}
